﻿/*	キャラクタLCD(HD44780) フォントデータ作成ツール
	『昼夜逆転』工作室	http://jsdiy.web.fc2.com

【更新履歴】
2013/10/19	ver 1.00	初版
*/

using System;
using System.Drawing;
using System.IO;
using System.Text;

namespace CharLCDFontCreator
{
	class Program
	{
		static void Main(string[] args)
		{
			if (args.Length == 0 || !File.Exists(args[0]))
			{
				ShowHelp();
				AppClose();
			}

			CLFC clfc = new CLFC();
			clfc.AppStart(args[0]);

			AppClose();
		}
		
		private static void ShowHelp()
		{
			Console.WriteLine(
				"Usage: >CharLCDFontCreator.exe charactertable.bmp" + Environment.NewLine +
				"出力をリダイレクトする場合は[Press any key]操作を忘れずに."
				);
		}

		public static void AppClose()
		{
			Console.WriteLine("Press any key.");
			Console.ReadKey();
			Environment.Exit(0);
		}

	}	//class

	class CLFC
	{
		/*
		pdf:サイズ100％でキャプチャした画像を読み込む。
		文字コード0x00のカラムの左上（枠線を含む）を原点(0,0)として画像（文字コード表）をトリミングしておく。
		カラムの大きさが縦横とも不揃いなので、画像編集してカラムをすべて同じ形（大きさ）にする。
		カラム内の文字（ドット表現された文字）の位置とドットの形は特にいじらなくてよい。
		*/
		private const int
			ColumnStepX = 37, ColumnStepY = 47,	//次のカラムまでの距離（1カラムの大きさ）
			GlyphCenterX = 18, GlyphCenterY = 16,	//文字形の中心位置（カラム左上からの距離）
			TotalDotW = 23, TotalDotH = 33,	//文字部分のドット合計の長さ
			NumOfDotX = 5, NumOfDotY = 7,	//文字部分のドット数
			NumOfCharX = 12, NumOfCharY = 16;	//文字コード0x20以上0xE0未満の文字の個数（文字コード表の文字の並びで）

		private const double
			DotW = (double)TotalDotW / NumOfDotX, DotH = (double)TotalDotH / NumOfDotY,	//ドットの大きさ
			ToStartDotX = DotW * (int)(NumOfDotX / 2), ToStartDotY = DotH * (int)(NumOfDotY / 2);	//グリフ中心からグリフ端のドットまでの距離
		/*
		ToStartDotXYは、ドットが奇数個ならグリフ端のドットの中心を指すことになる。
		なぜこのパラメータが必要か？
		ドットの形がいびつなのでドットを移動する（スキャンする）ごとに誤差が累積され、後になるほどサンプリング精度が落ちる。
		グリフ中心のドットからグリフ端のドット（スキャン開始位置）を割り出せば、誤差の影響を抑えることができる。
		*/

		private Bitmap bmp = null;

		public void AppStart(string filePath)
		{
			try { bmp = new Bitmap(filePath); }
			catch
			{
				Console.WriteLine("'{0}'がbmpまたはpngではありません.", filePath);
				Program.AppClose();
			}

			int[][] fontValues = new int[NumOfCharX * NumOfCharY][];
			int charCount = 0;
			for (int charCountX = 0; charCountX < NumOfCharX; charCountX++)
			{
				int columnX = charCountX * ColumnStepX;
				for (int charCountY = 0; charCountY < NumOfCharY; charCountY++)
				{
					int columnY = charCountY * ColumnStepY;
					fontValues[charCount++] = DotSampling(columnX, columnY);
				}
			}
			bmp.Dispose();

			PrintOut(fontValues);
		}

		private int[] DotSampling(int columnX, int columnY)
		{
			int[] fontData = new int[NumOfDotX];
			byte whiteLowLimit = 0x30;	//白と判断する色成分の閾値（これより大きい値は白と見なす）

			for (int dotCountX = 0; dotCountX < NumOfDotX; dotCountX++)
			{
				double dotX = (columnX + GlyphCenterX - ToStartDotX) + DotW * dotCountX;
				int fontDataV = 0x00;	//縦1ラインのフォントデータ
				for (int dotCountY = 0; dotCountY < NumOfDotY; dotCountY++)
				{
					double dotY = (columnY + GlyphCenterY - ToStartDotY) + DotH * dotCountY;
					int x = (int)(dotX + 0.5);	//四捨五入
					int y = (int)(dotY + 0.5);
					Color color = bmp.GetPixel(x, y);
					int bit = (whiteLowLimit < (color.R + color.G + color.B) / 3) ? 0 : 1;
					fontDataV |= (bit << dotCountY);
				}
				fontData[dotCountX] = fontDataV;
			}

			return fontData;
		}

		private void PrintOut(int[][] fontValues)
		{
			Encoding sjis = Encoding.GetEncoding("shift_jis");
			StringBuilder sb = new StringBuilder();

			Console.WriteLine("{");

			//「	{ 0x00, 0x00, 0x00, 0x00, 0x00 },	// 0x41 A」という書式
			for (int cc = 0; cc < fontValues.Length; cc++)
			{
				sb = sb.Clear();

				sb = sb.Append("\t{ ");
				int nullCheck = 0x00;
				for (int v = 0; v < fontValues[cc].Length; v++)
				{
					if (0 < v) sb = sb.Append(", ");
					sb = sb.AppendFormat("0x{0:X2}", fontValues[cc][v]);
					nullCheck |= fontValues[cc][v];
				}
				sb = sb.Append(" }");
				if (cc != fontValues.Length - 1) sb = sb.Append(",");

				byte charCode = (byte)(0x20 + cc);
				string s = sjis.GetString(new byte[] { charCode });
				if (charCode == 0x20) s = @"' '";	//スペース文字
				if (charCode == 0x5C) s = @"'\'";	//エスケープ文字
				if (charCode == 0x7E) s = @"'->'";	//Shift-JIS文字コードと違う文字
				if (charCode == 0x7F) s = @"'<-'";	//Shift-JIS文字コードと違う文字
				if (nullCheck == 0x00 && charCode != 0x20) s = "";
				sb = sb.AppendFormat("\t// 0x{0:X2} {1}", charCode, s);

				string line = sb.ToString();
				Console.WriteLine(line);
			}

			Console.WriteLine("};\t//HD44780 Character Patterns (ROM Code:A00) Shift-JIS");
		}

	}	//class
}	//namespace
